from cat_ui.containers import BoxContainer, VerticalList, HorizontalList, Alignment
from cat_ui.visual import Label
from cat_ui.styles import normal, bold_normal


class Button(BoxContainer):
    def __init__(self, text: str="", min_width=3, min_height=3, left_padding=1, right_padding=1, top_padding=1, bottom_padding=1, alignment=Alignment.TOP_LEFT, text_alignment=Alignment.CENTER, style=normal, selected_style=bold_normal):
        super().__init__(min_width, min_height, left_padding, right_padding, top_padding, bottom_padding, alignment, **style)
        self.children: list[Label] = [Label(text, alignment=text_alignment)]
        self.actions = []
        self.style = style
        self.text = text
        self.selected_style = selected_style
    
    def append(self, child):
        raise NotImplementedError()
    
    def process_key(self, raw_key):
        if raw_key in "\n\r":
            self.on_click()
        return raw_key
    
    def set_text(self, text):
        self.text = text
        self.children[0].set_text(text)
    
    def get_text(self):
        return self.text

    def select(self):
        self.set_style(**self.selected_style)
        self.children[0].set_text(f"\033[1m{self.text}\033[0;0m")
    
    def deselect(self):
        self.set_style(**self.style)
        self.children[0].set_text(self.text)
    
    def add_action(self, on_click):
        self.actions.append(on_click)

    def on_click(self):
        for action in self.actions:
            action(self)
    
    def selectable(self):
        return True


class ButtonText(Label):
    def __init__(self, text, width=-1, alignment=Alignment.TOP_LEFT):
        super().__init__(text, width, alignment)
        self.actions = []
        self.plain_text = text
        self.selected = False
    
    def process_key(self, raw_key):
        if raw_key in "\n\r":
            self.on_click()
        return raw_key
    
    def set_text(self, text):
        self.plain_text = text
        self.update_text()
    
    def get_text(self):
        return self.plain_text

    def select(self):
        self.selected = True
        self.update_text()
    
    def deselect(self):
        self.selected = False
        self.update_text()
    
    def update_text(self):
        super().set_text(f"\033[1m{self.plain_text}\033[0;0m" if self.selected else self.plain_text)
    
    def add_action(self, on_click):
        self.actions.append(on_click)

    def on_click(self):
        for action in self.actions:
            action(self)
    
    def selectable(self):
        return True


class SelectorLabel(Label):
    def __init__(self, text, chosen_prefix="*", plain_prefix=" ", width=-1, alignment=Alignment.TOP_LEFT):
        super().__init__(plain_prefix + text, width, alignment)
        self.chosen_prefix = chosen_prefix
        self.plain_prefix = plain_prefix
        self.plain_text = text
        self.selected = False
        self.chosen = False
        self.actions_on = []
        self.actions_off = []

    def set_text(self, text: str, width = -1):
        self.plain_text = text
        self.update_text()
    
    def select(self):
        self.selected = True
        self.update_text()
    
    def deselect(self):
        self.selected = False
        self.update_text()
    
    def update_text(self):
        text_with_prefix = (self.chosen_prefix if self.chosen else self.plain_prefix) + self.plain_text
        super().set_text(f"\033[1m{text_with_prefix}\033[0;0m" if self.selected else text_with_prefix)
    
    def add_action_on(self, on_click):
        self.actions_on.append(on_click)
        
    def add_action_off(self, on_click):
        self.actions_off.append(on_click)
    
    def chose(self, state: bool):
        self.chosen = state
        for action in self.actions_on if state else self.actions_off:
            action(self)
        self.update_text()
    
    def process_key(self, raw_key):
        if raw_key in '\n\r':
            self.chose(not self.chosen)
        return raw_key
    
    def selectable(self):
        return True


class VerticalCheckbox(VerticalList):
    def __init__(self, options: list[str], min_width=3, min_height=3, space=1, alignment=Alignment.TOP_LEFT, text_alignment=Alignment.CENTER_LEFT, chosen_prefix="*", plain_prefix=" "):
        super().__init__(min_width, min_height, space, alignment)
        self.children: list[SelectorLabel] = [SelectorLabel(option, chosen_prefix, plain_prefix, alignment=text_alignment) for option in options]
        self.actions = []
    
    def crop(self, max_width):
        for child in self.children:
            child.crop(max_width)

    def expand(self):
        for child in self.children:
            child.expand()
    
    def get_mask(self):
        return [child.chosen for child in self.children]
    
    def get_indices(self):
        indices = []
        for index, child in enumerate(self.children):
            if child.chosen:
                indices.append(index)
        return indices
    
    def process_key(self, raw_key):
        raw_key = super().process_key(raw_key)
        if raw_key in '\n\r':
            self.on_change()
        return raw_key
    
    def add_action(self, on_click):
        self.actions.append(on_click)

    def on_change(self):
        for action in self.actions:
            action(self)
    
    def selectable(self):
        return bool(self.children)

class HorizontalCheckbox(HorizontalList):
    def __init__(self, options: list[str], min_width=3, min_height=3, space=1, alignment=Alignment.TOP_LEFT, text_alignment=Alignment.CENTER_LEFT, chosen_prefix="*", plain_prefix=" "):
        super().__init__(min_width, min_height, space, alignment)
        self.children: list[SelectorLabel] = [SelectorLabel(option, chosen_prefix, plain_prefix, alignment=text_alignment) for option in options]
        self.actions = []
    
    def crop(self, max_width):
        for child in self.children:
            child.crop(max_width)

    def expand(self):
        for child in self.children:
            child.expand()
    
    def get_mask(self):
        return [child.chosen for child in self.children]
    
    def get_indices(self):
        indices = []
        for index, child in enumerate(self.children):
            if child.chosen:
                indices.append(index)
        return indices
    
    def process_key(self, raw_key):
        raw_key = super().process_key(raw_key)
        if raw_key in '\n\r':
            self.on_change()
        return raw_key
    
    def add_action(self, on_click):
        self.actions.append(on_click)

    def on_change(self):
        for action in self.actions:
            action(self)
    
    def selectable(self):
        return bool(self.children)


class VerticalRadio(VerticalList):
    def __init__(self, options: list[str], min_width=3, min_height=3, space=1, alignment=Alignment.TOP_LEFT, text_alignment=Alignment.CENTER_LEFT, chosen_prefix="*", plain_prefix=" "):
        super().__init__(min_width, min_height, space, alignment)
        self.children: list[SelectorLabel] = [SelectorLabel(option, chosen_prefix, plain_prefix, alignment=text_alignment) for option in options]
        def deselect_everyone_except(radio):
            for child in self.children:
                if child != radio:
                    child.chose(False)
        for child in self.children:
            child.add_action_on(deselect_everyone_except)
        self.actions = []
    
    def crop(self, max_width):
        for child in self.children:
            child.crop(max_width)

    def expand(self):
        for child in self.children:
            child.expand()
    
    def get_index(self):
        for index, child in enumerate(self.children):
            if child.chosen:
                return index
        return -1
    
    def process_key(self, raw_key):
        raw_key = super().process_key(raw_key)
        if raw_key in '\n\r':
            self.on_change()
        return raw_key
    
    def add_action(self, on_click):
        self.actions.append(on_click)

    def on_change(self):
        for action in self.actions:
            action(self)

    def selectable(self):
        return bool(self.children)

class HorizontalRadio(HorizontalList):
    def __init__(self, options: list[str], min_width=3, min_height=3, space=1, alignment=Alignment.TOP_LEFT, text_alignment=Alignment.CENTER_LEFT, chosen_prefix="*", plain_prefix=" "):
        super().__init__(min_width, min_height, space, alignment)
        self.children: list[SelectorLabel] = [SelectorLabel(option, chosen_prefix, plain_prefix, alignment=text_alignment) for option in options]
        def deselect_everyone_except(radio):
            for child in self.children:
                if child != radio:
                    child.chose(False)
        for child in self.children:
            child.add_action_on(deselect_everyone_except)
        self.actions = []
    
    def crop(self, max_width):
        for child in self.children:
            child.crop(max_width)

    def expand(self):
        for child in self.children:
            child.expand()
    
    def get_index(self):
        for index, child in enumerate(self.children):
            if child.chosen:
                return index
        return -1
    
    def process_key(self, raw_key):
        raw_key = super().process_key(raw_key)
        if raw_key in '\n\r':
            self.on_change()
        return raw_key
    
    def add_action(self, on_click):
        self.actions.append(on_click)

    def on_change(self):
        for action in self.actions:
            action(self)

    def selectable(self):
        return bool(self.children)



import sys
import termios
import tty
import os
from enum import Enum
import string
import regex


ansi_pattern = regex.compile(r'(\x1B\[[0-?]*[ -/]*[@-~])')
printable = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZабвгдежзийклмнопрстуфхцчшщъыьэюяАБВГДЕЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯ!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~ \t\n\r\x0b\x0c'


class Key(Enum):
    LOWER_A = "a"
    LOWER_B = "b"
    LOWER_C = "c"
    LOWER_D = "d"
    LOWER_E = "e"
    LOWER_F = "f"
    LOWER_G = "g"
    LOWER_H = "h"
    LOWER_I = "i"
    LOWER_J = "j"
    LOWER_K = "k"
    LOWER_L = "l"
    LOWER_M = "m"
    LOWER_N = "n"
    LOWER_O = "o"
    LOWER_P = "p"
    LOWER_Q = "q"
    LOWER_R = "r"
    LOWER_S = "s"
    LOWER_T = "t"
    LOWER_U = "u"
    LOWER_V = "v"
    LOWER_W = "w"
    LOWER_X = "x"
    LOWER_Y = "y"
    LOWER_Z = "z"
    A = "A"
    B = "B"
    C = "C"
    D = "D"
    E = "E"
    F = "F"
    G = "G"
    H = "H"
    I = "I"
    J = "J"
    K = "K"
    L = "L"
    M = "M"
    N = "N"
    O = "O"
    P = "P"
    Q = "Q"
    R = "R"
    S = "S"
    T = "T"
    U = "U"
    V = "V"
    W = "W"
    X = "X"
    Y = "Y"
    Z = "Z"
    NUMBER_0 = 0
    NUMBER_1 = 1
    NUMBER_2 = 2
    NUMBER_3 = 3
    NUMBER_4 = 4
    NUMBER_5 = 5
    NUMBER_6 = 6
    NUMBER_7 = 7
    NUMBER_8 = 8
    NUMBER_9 = 9
    UP_ARROW = "up"
    DOWN_ARROW = "down"
    RIGHT_ARROW = "right"
    LEFT_ARROW = "left"
    SHIFT = "shift"
    CTRL = "ctrl"
    ENTER = "enter"
    BACKSPASE = "backspace"
    OTHER = "other"


class KeyType(Enum):
    LETTER = 0
    NUMBER = 1
    ARROW = 2
    COMBINATION = 3
    OTHER = 4


def move_left(N):
    return f"\033[{N}D" if N > 0 else ""

def move_right(N):
    return f"\033[{N}C" if N > 0 else ""

def move_up(N):
    return f"\033[{N}A" if N > 0 else ""

def move_down(N):
    return f"\033[{N}B" if N > 0 else ""


def get_key_raw():
    fd = sys.stdin.fileno()
    old = termios.tcgetattr(fd)
    try:
        tty.setraw(fd)
        ch = sys.stdin.read(1)
        if ch == '\x1b':
            ch += sys.stdin.read(2)
            if ch == '\x1b[1':
                ch += sys.stdin.read(3)
        return ch
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old)


def get_key_from_raw(key):
    key_to_arrow = {"A": Key.UP_ARROW, "B": Key.DOWN_ARROW, "C": Key.RIGHT_ARROW, "D": Key.LEFT_ARROW}
    key_to_special = {"2": Key.SHIFT, "5": Key.CTRL}

    if len(key) == 1:
        if key in string.ascii_letters:
            return Key(key), KeyType.LETTER
        if key in [str(num) for num in range(10)]:
            return Key(int(key)), KeyType.NUMBER
        if ord(key) == 13:
            return Key.ENTER, KeyType.OTHER
        if 1 <= ord(key) <= 26:
            return (Key.CTRL, Key(chr(ord(key) + 64))), KeyType.COMBINATION
        if key == "\x7f":
            return Key.BACKSPASE, KeyType.OTHER
        return Key.OTHER, KeyType.OTHER
    elif len(key) == 3:
        return key_to_arrow[key[-1]], KeyType.ARROW
    elif len(key) == 6:
        return (key_to_special[key[-2]], key_to_arrow[key[-1]]), KeyType.COMBINATION
    return Key.OTHER, KeyType.OTHER


def visible_length(s):
    clean = ansi_pattern.sub('', s)
    return len(clean)


def split_visible_chars(s):
        parts = []
        current_prefix = ''
        current_suffix = ''
        i = 0
        while i < len(s):
            match = ansi_pattern.match(s, i)
            if match:
                code = match.group(1)
                if code.endswith('m'):
                    if code == '\033[0m':
                        current_suffix += code
                    else:
                        current_prefix += code
                i = match.end()
            else:
                visible = s[i]
                fragment = current_prefix + visible + current_suffix
                parts.append(fragment)
                i += 1
                current_suffix = ''
        return parts


def get_raw_from_key(key: Key):
    if key == Key.UP_ARROW:
        return "\x1b[A"
    if key == Key.DOWN_ARROW:
        return "\x1b[B"
    if key == Key.RIGHT_ARROW:
        return "\x1b[C"
    if key == Key.LEFT_ARROW:
        return "\x1b[D"
    if key.value in string.ascii_letters:
        return key.value
    if key.value in range(10):
        return str(key.value)
    if key.value == Key.ENTER:
        return chr(13)
    raise KeyError()


def get_key():
    key = get_key_raw()
    return get_key_from_raw(key)


def clear():
    os.system('clear')



from cat_ui.containers import ContainerElement, Alignment
from cat_ui.keyboard_manager import visible_length, split_visible_chars


class Label(ContainerElement):
    def __init__(self, text: str, width = -1, alignment = Alignment.TOP_LEFT):
        width = len(text) if width == -1 else width
        super().__init__(width, (len(text) + width - 1) // width if width > 0 else 0, alignment)
        self.text = text
    
    def __len__(self):
        return visible_length(self.text)

    def crop(self, max_width):
        self.width = min(len(self), max_width)
        self.height = (len(self) + self.width - 1) // self.width

    def expand(self):
        self.width = len(self)
        self.height = 1
    
    def set_text(self, text: str, width = -1):
        self.text = text
        blocks = self.text.split("\n")
        max_width = 0
        for block in blocks:
            max_width = max(visible_length(text) if width == -1 else width, max_width)
        self.width = max_width
        self.height = 0
        if self.width == 0:
            return
        for block in blocks:
            self.height += (visible_length(block) + self.width - 1) // self.width
    
    def get_text(self):
        return self.text

    def draw(self):
        text = f"\033[1B\033[{self.width}D".join(self.get_splited())
        return f"\033[s{text}\033[u"
    
    def get_splited(self):
        if self.width == 0:
            return []
        blocks = self.text.split("\n")
        splited = []
        for block in blocks:
            symbols = split_visible_chars(block)
            line_height = (visible_length(block) + self.width - 1) // self.width
            for line in range(line_height):
                line_content = "".join(symbols[self.get_width() * line: self.get_width() * (line + 1)])
                if visible_length(line_content) < self.get_width():
                    line_content += f"\033[{self.get_width() - visible_length(line_content)}C"
                splited.append(line_content)
        return splited



from cat_ui.keyboard_manager import move_left, move_right, move_up, move_down, get_key_from_raw, get_raw_from_key, Key, KeyType
from enum import Enum


class Alignment(Enum):
    TOP_LEFT = (0, 0)
    TOP_CENTER = (0, 0.5)
    TOP_RIGHT = (0, 1)
    CENTER_LEFT = (0.5, 0)
    CENTER = (0.5, 0.5)
    CENTER_RIGHT = (0.5, 1)
    BOTTOM_LEFT = (1, 0)
    BOTTOM_CENTER = (1, 0.5)
    BOTTOM_RIGHT = (1, 1)


class ContainerElement:
    def __init__(self, width, height, alignment: Alignment=Alignment.TOP_LEFT):
        self.width = width
        self.height = height
        self.alignment = alignment

    def draw(self):
        return ""

    def get_width(self):
        return self.width
    
    def get_height(self):
        return self.height
    
    def draw_on_screen(self, screen_width=0, screen_height=0):
        horizontal_alignment = "{context}"
        if int((screen_width - self.get_width())*self.alignment.value[1]) > 0:
            horizontal_alignment = f"\033[{int((screen_width - self.get_width())*self.alignment.value[1])}C{{context}}\033[{int((screen_width - self.get_width())*self.alignment.value[1])}D"
        vertical_alignment = "{context}"
        if int((screen_height - self.get_height())*self.alignment.value[0]) > 0:
            vertical_alignment = f"\033[{int((screen_height - self.get_height())*self.alignment.value[0])}B{{context}}\033[{int((screen_height - self.get_height())*self.alignment.value[0])}A"
        return vertical_alignment.format(context=horizontal_alignment.format(context=self.draw()))
    
    def process_key(self, raw_key):
        return raw_key
    
    def select(self):
        pass
    
    def deselect(self):
        pass

    def selectable(self):
        return False


class Box(ContainerElement):
    def __init__(self, width, height, alignment: Alignment=Alignment.TOP_LEFT, **style):
        super().__init__(width, height, alignment)
        self.set_style(**style)

    def set_style(self, vertical_bar="|", horizontal_bar="-", corners=["*", "*", "*", "*"], fill=" "):
        self.vertical_bar = vertical_bar
        self.horizonal_bar = horizontal_bar
        self.corners = corners
        self.fill = fill

    def draw(self):
        text = f"{self.vertical_bar}{self.fill*(self.get_width()-2)}{self.vertical_bar}{move_left(self.get_width())}\033[1B"*(self.get_height()-2)
        return f"\033[s{self.corners[0]}{self.horizonal_bar*(self.get_width() - 2)}{self.corners[1]}{move_left(self.get_width())}\033[1B{text}{self.corners[2]}{self.horizonal_bar*(self.get_width() - 2)}{self.corners[3]}\033[u"


class BoxContainer(Box):
    def __init__(self, min_width=3, min_height=3, left_padding=1, right_padding=1, top_padding=1, bottom_padding=1, alignment: Alignment=Alignment.TOP_LEFT, **style):
        super().__init__(min_width, min_height, alignment, **style)
        self.left_padding = left_padding
        self.right_padding = right_padding
        self.top_padding = top_padding
        self.bottom_padding = bottom_padding
        self.children: list[ContainerElement] = []
        self.selection_index = 0
    
    def append(self, child: ContainerElement):
        self.children.append(child)
    
    def get_width(self):
        return max(self.width, self.left_padding + self.right_padding + (max([child.get_width() for child in self.children]) if self.children else 0))
    
    def get_height(self):
        return max(self.height, self.top_padding + self.bottom_padding + (max([child.get_height() for child in self.children]) if self.children else 0))
    
    def draw(self):
        children_draw = "".join([child.draw_on_screen(self.get_width() - self.left_padding - self.right_padding, self.get_height() - self.top_padding - self.bottom_padding) for child in self.children])
        return f"{super().draw()}{move_right(self.left_padding)}{move_down(self.top_padding)}{children_draw}{move_up(self.left_padding)}{move_left(self.left_padding)}"
    
    def select(self):
        if not self.children[self.selection_index].selectable():
            self.selection_index = 0
            while not self.children[self.selection_index].selectable():
                self.selection_index += 1
                if self.selection_index == len(self.children):
                    return
        self.children[self.selection_index].select()
    
    def deselect(self):
        self.children[self.selection_index].deselect()
    
    def process_key(self, raw_key):
        if not self.children:
            return raw_key
        raw_key = self.children[self.selection_index].process_key(raw_key)
        key, key_type = get_key_from_raw(raw_key)
        if key_type == KeyType.COMBINATION:
            return get_raw_from_key(key[-1])
        if key_type == KeyType.ARROW:
            if key in [Key.UP_ARROW, Key.LEFT_ARROW]:
                new_selection = self.selection_index - 1
                found_selection = False
                while new_selection >= 0:
                    if self.children[new_selection].selectable():
                        found_selection = True
                        break
                    new_selection -= 1
                if found_selection:
                    self.children[self.selection_index].deselect()
                    self.selection_index = new_selection
                    self.children[self.selection_index].select()
                    return ""
                return raw_key
            new_selection = self.selection_index + 1
            found_selection = False
            while new_selection < len(self.children):
                if self.children[new_selection].selectable():
                    found_selection = True
                    break
                new_selection += 1
            if found_selection:
                self.children[self.selection_index].deselect()
                self.selection_index = new_selection
                self.children[self.selection_index].select()
                return ""
            return raw_key
        return raw_key
    
    def selectable(self):
        return any([child.selectable() for child in self.children])


class VerticalList(ContainerElement):
    def __init__(self, min_width=3, min_height=3, space=1, alignment: Alignment=Alignment.TOP_LEFT):
        super().__init__(min_width, min_height, alignment)
        self.space = space
        self.children: list[ContainerElement] = []
        self.selection_index = 0
    
    def append(self, child: ContainerElement):
        self.children.append(child)
    
    def get_width(self):
        return max(self.width, max([child.get_width() for child in self.children]) if self.children else 0)
    
    def get_height(self):
        return max(self.height, (sum([child.get_height() + self.space for child in self.children]) if self.children else 0) - self.space)
    
    def draw(self):
        total_movement = sum([child.get_height() + self.space for child in self.children]) - self.space
        children_draw = move_down(self.space).join([f"{child.draw_on_screen(screen_width=self.get_width())}{move_down(child.get_height())}" for child in self.children])
        return f"{children_draw}{move_up(total_movement)}"
    
    def select(self):
        if not self.children[self.selection_index].selectable():
            self.selection_index = 0
            while not self.children[self.selection_index].selectable():
                self.selection_index += 1
                if self.selection_index == len(self.children):
                    return
        self.children[self.selection_index].select()
    
    def deselect(self):
        self.children[self.selection_index].deselect()
    
    def process_key(self, raw_key):
        if not self.children:
            return raw_key
        raw_key = self.children[self.selection_index].process_key(raw_key)
        key, key_type = get_key_from_raw(raw_key)
        if key_type == KeyType.COMBINATION:
            return get_raw_from_key(key[-1])
        if key_type == KeyType.ARROW:
            if key == Key.UP_ARROW:
                new_selection = self.selection_index - 1
                found_selection = False
                while new_selection >= 0:
                    if self.children[new_selection].selectable():
                        found_selection = True
                        break
                    new_selection -= 1
                if found_selection:
                    self.children[self.selection_index].deselect()
                    self.selection_index = new_selection
                    self.children[self.selection_index].select()
                    return ""
                return raw_key
            if key == Key.DOWN_ARROW:
                new_selection = self.selection_index + 1
                found_selection = False
                while new_selection < len(self.children):
                    if self.children[new_selection].selectable():
                        found_selection = True
                        break
                    new_selection += 1
                if found_selection:
                    self.children[self.selection_index].deselect()
                    self.selection_index = new_selection
                    self.children[self.selection_index].select()
                    return ""
                return raw_key
        return raw_key
    
    def selectable(self):
        return any([child.selectable() for child in self.children])


class HorizontalList(ContainerElement):
    def __init__(self, min_width=3, min_height=3, space=1, alignment: Alignment=Alignment.TOP_LEFT):
        super().__init__(min_width, min_height, alignment)
        self.space = space
        self.children: list[ContainerElement] = []
        self.selection_index = 0
    
    def append(self, child: ContainerElement):
        self.children.append(child)
    
    def get_width(self):
        return max(self.width, (sum([child.get_width() + self.space for child in self.children]) if self.children else 0) - self.space)
    
    def get_height(self):
        return max(self.height, max([child.get_height() for child in self.children]) if self.children else 0)
    
    def draw(self):
        total_movement = sum([child.get_width() + self.space for child in self.children]) - self.space
        children_draw = move_right(self.space).join([f"{child.draw_on_screen(screen_height=self.get_height())}{move_right(child.get_width())}" for child in self.children])
        return f"{children_draw}{move_left(total_movement)}"
    
    def select(self):
        if not self.children[self.selection_index].selectable():
            self.selection_index = 0
            while not self.children[self.selection_index].selectable():
                self.selection_index += 1
                if self.selection_index == len(self.children):
                    return
        self.children[self.selection_index].select()
    
    def deselect(self):
        self.children[self.selection_index].deselect()
    
    def process_key(self, raw_key):
        if not self.children:
            return raw_key
        raw_key = self.children[self.selection_index].process_key(raw_key)
        key, key_type = get_key_from_raw(raw_key)
        if key_type == KeyType.COMBINATION:
            return get_raw_from_key(key[-1])
        if key_type == KeyType.ARROW:
            if key == Key.LEFT_ARROW:
                new_selection = self.selection_index - 1
                found_selection = False
                while new_selection >= 0:
                    if self.children[new_selection].selectable():
                        found_selection = True
                        break
                    new_selection -= 1
                if found_selection:
                    self.children[self.selection_index].deselect()
                    self.selection_index = new_selection
                    self.children[self.selection_index].select()
                    return ""
                return raw_key
            if key == Key.RIGHT_ARROW:
                new_selection = self.selection_index + 1
                found_selection = False
                while new_selection < len(self.children):
                    if self.children[new_selection].selectable():
                        found_selection = True
                        break
                    new_selection += 1
                if found_selection:
                    self.children[self.selection_index].deselect()
                    self.selection_index = new_selection
                    self.children[self.selection_index].select()
                    return ""
                return raw_key
        return raw_key
    
    def selectable(self):
        return any([child.selectable() for child in self.children])


class VerticalContainer(BoxContainer):
    def __init__(self, min_width=3, min_height=3, left_padding=1, right_padding=1, top_padding=1, bottom_padding=1, space=1, alignment = Alignment.TOP_LEFT, **style):
        super().__init__(min_width, min_height, left_padding, right_padding, top_padding, bottom_padding, alignment, **style)
        self.children = [VerticalList(min_width - left_padding - right_padding, min_height - top_padding - bottom_padding, space)]
    
    def append(self, child: ContainerElement):
        self.children[0].append(child)


class HorizontalContainer(BoxContainer):
    def __init__(self, min_width=3, min_height=3, left_padding=1, right_padding=1, top_padding=1, bottom_padding=1, space=1, alignment = Alignment.TOP_LEFT, **style):
        super().__init__(min_width, min_height, left_padding, right_padding, top_padding, bottom_padding, alignment, **style)
        self.children = [HorizontalList(min_width - left_padding - right_padding, min_height - top_padding - bottom_padding, space)]
    
    def append(self, child: ContainerElement):
        self.children[0].append(child)



from setuptools import setup, find_packages

setup(
    name="cat_ui",
    version="1.1.1",
    description="Simple python ui lib for cli programs",
    author="Akio Smiowly",
    author_email="lazzycatto@yandex.com",
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "regex>=2024.11.6",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
)



from cat_ui.containers import ContainerElement, BoxContainer, Alignment
from cat_ui.keyboard_manager import clear, get_key_raw, get_key_from_raw, Key, KeyType


class App():
    def __init__(self, width=100, height=50, main_window: ContainerElement=BoxContainer(alignment=Alignment.CENTER)):
        self.width = width
        self.height = height
        self.main_window = main_window
        self.running = False

    async def run(self):
        self.running = True
        clear()
        print(self.main_window.draw_on_screen(self.width, self.height), end="", flush=True)
        raw_key = get_key_raw()
        key, key_type = get_key_from_raw(raw_key)
        while True:
            if key_type == KeyType.COMBINATION and key[-1] == Key.C:
                break
            self.main_window.process_key(raw_key)
            if not self.running:
                break
            clear()
            print(self.main_window.draw_on_screen(self.width, self.height), end="", flush=True)
            raw_key = get_key_raw()
            key, key_type = get_key_from_raw(raw_key)

        clear()
        self.running = False
    
    def set_screen(self, window: ContainerElement):
        self.main_window = window
        self.main_window.select()
        if self.running:
            clear()
            print(self.main_window.draw_on_screen(self.width, self.height), end="", flush=True)
    
    def stop(self):
        self.running = False



from cat_ui.containers import Alignment
from cat_ui.visual import Label
from cat_ui.keyboard_manager import get_key_from_raw, get_raw_from_key, Key, KeyType, printable


class InputField(Label):
    def __init__(self, text="", width=-1, max_length=-1, alignment=Alignment.TOP_LEFT, alowed_char=printable):
        super().__init__(text, width, alignment)
        self.alowed_char = alowed_char
        self.max_length = max_length
        self.selected = False
        self.selected_character = len(text)
        self.plane_text = text
        self.max_width = width
        self.actions = []

    def set_width(self, width):
        self.max_width = width
        self.update_label()
    
    def select(self):
        self.selected = True
        self.update_label()
    
    def deselect(self):
        self.selected = False
        self.update_label()
    
    def process_key(self, raw_key):
        if raw_key in "\n\r":
            self.on_enter()
            return ""
        if raw_key in self.alowed_char and (self.max_length == -1 or len(self.plane_text) < self.max_length):
            self.plane_text = self.plane_text[:self.selected_character] + raw_key + self.plane_text[self.selected_character:]
            self.selected_character += 1
            self.update_label()
        key, key_type = get_key_from_raw(raw_key)
        key, key_type = get_key_from_raw(raw_key)
        if key_type == KeyType.COMBINATION:
            return get_raw_from_key(key[-1])
        if key_type == KeyType.ARROW:
            if key == Key.LEFT_ARROW:
                if self.selected_character > 0:
                    self.selected_character -= 1
                    self.update_label()
                    return ""
                return raw_key
            if key == Key.RIGHT_ARROW:
                if self.selected_character < len(self.plane_text):
                    self.selected_character += 1
                    self.update_label()
                    return ""
                return raw_key
        if key_type == KeyType.OTHER:
            if key == Key.BACKSPASE:
                if self.selected_character > 0:
                    self.plane_text = self.plane_text[:self.selected_character - 1] + self.plane_text[self.selected_character:]
                    self.selected_character -= 1
                    self.update_label()
        return raw_key
    
    def set_text(self, text, width=-1):
        self.plane_text = text
        self.max_width = width
        self.update_label()
    
    def get_text(self):
        return self.plane_text

    def add_action(self, on_click):
        self.actions.append(on_click)

    def on_enter(self):
        for action in self.actions:
            action(self)

    def update_label(self):
        if self.selected:
            if self.selected_character < len(self.plane_text):
                super().set_text(f"\033[1m{self.plane_text[:self.selected_character]}\033[4m{self.plane_text[self.selected_character]}\033[0;0m\033[1m{self.plane_text[self.selected_character + 1:]}\033[0;0m", self.max_width)
            else:
                super().set_text(f"\033[1m{self.plane_text}_\033[0;0m", self.max_width)
        else:
            super().set_text(self.plane_text, self.max_width)
    
    def selectable(self):
        return True


class PasswordInput(Label):
    def __init__(self, text="", password_char="*", width=-1, max_length=-1, alignment=Alignment.TOP_LEFT, alowed_char=printable):
        super().__init__(text, width, alignment)
        self.alowed_char = alowed_char
        self.max_length = max_length
        self.password_char = password_char
        self.selected = False
        self.selected_character = len(text)
        self.plane_text = text
        self.max_width = width
        self.actions = []

    def set_width(self, width):
        self.max_width = width
        self.update_label()
    
    def select(self):
        self.selected = True
        self.update_label()
    
    def deselect(self):
        self.selected = False
        self.update_label()
    
    def process_key(self, raw_key):
        if raw_key in "\n\r":
            self.on_enter()
            return ""
        if raw_key in self.alowed_char and (self.max_length == -1 or len(self.plane_text) < self.max_length):
            self.plane_text = self.plane_text[:self.selected_character] + raw_key + self.plane_text[self.selected_character:]
            self.selected_character += 1
            self.update_label()
        key, key_type = get_key_from_raw(raw_key)
        key, key_type = get_key_from_raw(raw_key)
        if key_type == KeyType.COMBINATION:
            return get_raw_from_key(key[-1])
        if key_type == KeyType.ARROW:
            if key == Key.LEFT_ARROW:
                if self.selected_character > 0:
                    self.selected_character -= 1
                    self.update_label()
                    return ""
                return raw_key
            if key == Key.RIGHT_ARROW:
                if self.selected_character < len(self.plane_text):
                    self.selected_character += 1
                    self.update_label()
                    return ""
                return raw_key
        if key_type == KeyType.OTHER:
            if key == Key.BACKSPASE:
                if self.selected_character > 0:
                    self.plane_text = self.plane_text[:self.selected_character - 1] + self.plane_text[self.selected_character:]
                    self.selected_character -= 1
                    self.update_label()
        return raw_key
    
    def set_text(self, text, width=-1):
        self.plane_text = text
        self.max_width = width
        self.update_label()
    
    def get_text(self):
        return self.plane_text

    def add_action(self, on_click):
        self.actions.append(on_click)

    def on_enter(self):
        for action in self.actions:
            action(self)

    def update_label(self):
        if self.selected:
            if self.selected_character < len(self.plane_text):
                super().set_text(f"\033[1m{self.password_char * self.selected_character}\033[4m{self.password_char}\033[0;0m\033[1m{self.password_char * (len(self.plane_text) -  self.selected_character - 1)}\033[0;0m", self.max_width)
            else:
                super().set_text(f"\033[1m{self.password_char * len(self.plane_text)}_\033[0;0m", self.max_width)
        else:
            super().set_text(self.password_char * len(self.plane_text), self.max_width)
    
    def selectable(self):
        return True



from cat_ui.containers import *
from cat_ui.keyboard_manager import *
from cat_ui.input_containers import *
from cat_ui.visual import *
from cat_ui.selectors import *
from cat_ui.app import *
import cat_ui.styles as styles

__all__ = [
    "App",
    "styles",
    "ansi_pattern",
    "printable",
    "Key",
    "KeyType",
    "move_left",
    "move_right",
    "move_up",
    "move_down",
    "get_key_raw",
    "get_key_from_raw",
    "visible_length",
    "split_visible_chars",
    "get_raw_from_key",
    "get_key",
    "clear",
    "Alignment",
    "ContainerElement",
    "Box",
    "BoxContainer",
    "VerticalList",
    "HorizontalList",
    "VerticalContainer",
    "HorizontalContainer",
    "Button",
    "ButtonText",
    "SelectorLabel",
    "VerticalCheckbox",
    "HorizontalCheckbox",
    "VerticalRadio",
    "HorizontalRadio",
    "Label",
    "InputField",
    "PasswordInput",
    "SCREEN_WIDTH",
    "SCREEN_HEIGHT",
    "run"
]
